function [psi_, ene_, conv_, err_] = EIGS(ham_, psi_, max_, niter_, thresh_)
  %% ham_: ham_(1) is the left part of effective Hamiltonian
  %%       ham_(2) is the central part of effective Hamiltonian
  %%       ham_(3) is the right part of effective Hamiltonian
  %% H(1)      H(2)       H(3)
  %% --[3]      [3]      [3]--
  %% |           |           |
  %% --[2]   [1]---[4]   [2]--
  %% |           |           |
  %% --[1]      [2]      [1]--
  %%
  %% psi_: single-site MPS
  %% [1]---[3] 
  %%     |
  %%    [2]
  %% you should not need to touch the rest...
  if isempty(max_), max_ = 50;, end;
  if isempty(niter_), niter_ = 1000;, end;
  if isempty(thresh_), thresh_ = 1e-8;, end;

  nstate = size(psi_,1);
  maxV = max_*nstate;
  conv_ = false(nstate,1);
  vU = []; vW = []; mH = []; mO = [];

  for iter = 1 : niter_
    nsize = size(psi_,2);
    new_vW = prod_HA(ham_, psi_);

    mH2 = prod_AB(psi_, new_vW);
    mO2 = prod_AB(psi_, psi_);
    if ~isempty(vU)
      mH1 = prod_AB(vU, new_vW);
      mO1 = prod_AB(vU, psi_);
      mH = [mH, mH1; mH1', mH2];
      mO = [mO, mO1; mO1', mO2];
    else
      mH = mH2;
      mO = mO2;
    end
    vU = [vU; psi_];
    vW = [vW; new_vW];
    
    %% Löwdin's orthogonalization
    [oV, oE] = eig(mO);
    oE = diag(oE);
    idx = find(oE > 1e-16);  
    oE = oE(idx);    
    oV = oV(:,idx);
    rotO = oV ./ sqrt(oE');
    %assert((size(rotO,2)>=nstate), 'guess vectors are linear dependent');

    %% diaganalize mH
    [eigV, mE] = eig(rotO' * mH * rotO);
    mE = diag(mE);
    [mE, ind] = sort(mE);
    eigV = eigV(:,ind);

    %% update vU & vW
    ene_ = mE(1:nstate);
    eigV = rotO * eigV(:,1:nstate);
    vU(1:nstate) = trans_A(vU,eigV);
    vW(1:nstate) = trans_A(vW,eigV);

    %% update mH & mO
    dimV = size(eigV,1);
    rotV = eye(dimV, dimV);
    rotV(:,1:nstate) = eigV;
    mH = rotV' * mH * rotV;
    mO = rotV' * mO * rotV;

    %% reduce size
    if (dimV>maxV)
      ns = dimV - nsize;
      remove = [];
      for ii = 1 : nstate
        if ~(conv_(ii))
          [~,idx] = mink(abs((eigV(nstate+1:ns,ii))), ii);
          for jj = 1 : ii
            id = any(remove==idx(jj));
            if ~id
              remove = [remove, idx(jj)];
              break;
            end
          end
        end
      end
      remove = remove + nstate;

      for ii = 1 : length(remove)
        vU(remove:end-1) = vU(remove+1:end); vU = vU(1:end-1);
        vW(remove:end-1) = vW(remove+1:end); vW = vW(1:end-1);
      end
      mH(:,remove) = []; mH(remove,:) = [];
      mO(:,remove) = []; mO(remove,:) = [];
    end
 
    %% compute residual
    [residual, err_] = compute_R(vW, vU, ene_);
    conv_ = (err_ < thresh_); %if (0~=1), [ene_; err]; ,end
    if (sum(conv_)~=nstate) && (iter~=niter_)
      idx = find(~conv_);
      psi_ = QSpace(length(idx));
      for ii = 1 : length(idx)
        psi_(ii) = residual(idx(ii))/err_(idx(ii));
      end
    else
      psi_ = vU(1:nstate); % check: vecnorm(ham_*psi_-psi_*diag(ene_))
      break;
    end
  end
end

function [out_] = prod_HA(ham_, v1_)
  n1 = size(v1_,1);
  out_ = QSpace(n1);
  for ii = 1 : n1
    tmp = contract(ham_(1), [3], v1_(ii), [1]);
    tmp = contract(tmp, [2,3], ham_(2), [1,3]);
    out_(ii) = contract(tmp, [4,2], ham_(3), [2,3]);
  end
end

function [out_] = prod_AB(v1_, v2_)
  n1 = size(v1_,1);
  n2 = size(v2_,1);
  out_ = zeros(n1, n2);
  for ii = 1 : n1
    for jj = 1 : n2
      tmp = contract(conj(v1_(ii)), [1,2,3], v2_(jj), [1,2,3]);
      if (length(tmp.data)~=0)
        out_(ii,jj) = tmp.data{1};
      end
    end
  end
end

function [out_] = trans_A(v1_, mat_)
  [n1,n2] = size(mat_);
  out_ = QSpace(n2);
  for ii = 1 : n2
    for jj = 1 : n1
      out_(ii) = out_(ii) + v1_(jj) * mat_(jj,ii);
    end
  end
end

function [out_, err_] = compute_R(vW, vU, ene_)
  %residual = vW(:,1:nstate) - vU(:,1:nstate) .* ene_';
  n1 = size(ene_,1);
  out_ = QSpace(n1);
  err_ = zeros(n1,1);
  for ii = 1 : n1
    out_(ii) = vW(ii) - vU(ii) * ene_(ii);
    err_(ii) = normQS(out_(ii));
  end
end
